@@ -1,22 +1,34 @@
-/* -*- linux-c -*-
+/* ----------------------------------------------------------------------- *
  *
- * automount.c
+ *  automount.c - Linux automounter daemon
+ *   
+ *   Copyright 1997 Transmeta Corporation - All Rights Reserved
  *
- *  Automount daemon for Linux autofs
+ *   This program is free software; you can redistribute it and/or modify
+ *   it under the terms of the GNU General Public License as published by
+ *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
+ *   USA; either version 2 of the License, or (at your option) any later
+ *   version.
+ *   
+ *   This program is distributed in the hope that it will be useful,
+ *   but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *   GNU General Public License for more details.
  *
- */
+ * ----------------------------------------------------------------------- */
 
 #include <dirent.h>
-#include <stdio.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <stdlib.h>
+#include <getopt.h>
+#include <limits.h>
 #include <paths.h>
 #include <signal.h>
+#include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
-#include <limits.h>
 #include <syslog.h>
+#include <unistd.h>
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -57,19 +69,6 @@ static struct autofs_point {
 #define DEFAULT_TIMEOUT (5*60)	/* 5 minutes */
 #define CHECK_RATIO     4	/* exp_runfreq = exp_timeout/CHECK_RATIO */
 
-/* Used by subprocesses to avoid doing silly things with signals */
-static void reset_signals(void)
-{
-	struct sigaction sa;
-	int i;
-
-	sa.sa_handler = SIG_DFL;
-	sigemptyset(&sa.sa_mask);
-	sa.sa_flags = 0;
-	for ( i = 1 ; i < NSIG ; i++ )
-		sigaction(i, &sa, NULL);
-}
-
 static int umount_ent(char *root, char *name)
 {
 	char path_buf[PATH_MAX];
@@ -218,15 +217,28 @@ static void run_expire(void)
 {
 	pid_t f;
 	struct autofs_packet_expire pkt;
+	sigset_t s, olds;
 
 	if ( ap.exp_process )	/* Another process is already expiring */
 		return;		/* Try again next interval */
+
+	if ( ioctl(ap.ioctlfd, AUTOFS_IOC_EXPIRE, &pkt) )
+		return;		/* Don't fork if there is nothing do expire */
 	
+	/* Temporarily block SIGCHLD and SIGALRM between forking and setting
+	   ap.exp_process */
+
+	sigemptyset(&s);
+	sigaddset(&s, SIGCHLD);
+	sigaddset(&s, SIGALRM);
+	sigprocmask(SIG_BLOCK, &s, &olds);
+
 	f = fork();
 	if ( f < 0 )
 		return;
 	if ( f > 0 ) {
 		ap.exp_process = f;
+		sigprocmask(SIG_SETMASK, &olds, NULL);
 		return;
 	}
 
@@ -237,13 +249,13 @@ static void run_expire(void)
 
 	syslog(LOG_DEBUG, "running expiration on path %s", ap.path);
 
-	while ( ioctl(ap.ioctlfd, AUTOFS_IOC_EXPIRE, &pkt) == 0 ) {
+	do {
 		if ( pkt.hdr.type == autofs_ptype_expire ) {
 			if ( !umount_ent(ap.path,pkt.name) )
 				syslog(LOG_NOTICE, "expired %s/%s", ap.path,
 				       pkt.name);
 		}
-	}
+	} while ( ioctl(ap.ioctlfd, AUTOFS_IOC_EXPIRE, &pkt) == 0 );
 	_exit(0);
 }
 
@@ -383,7 +395,7 @@ static int handle_packet(void)
 	return 0;
 }
 
-void sig_child(int sig)
+static void sig_child(int sig)
 {
 	pid_t pid;
 	int status;
@@ -412,7 +424,7 @@ void sig_child(int sig)
 	}
 }
 
-static void initialize(void)
+static void become_daemon(void)
 {
 	pid_t pid;
 
@@ -443,33 +455,90 @@ static void initialize(void)
 
 }
 
+static unsigned long getnumopt(char *str, char option)
+{
+  unsigned long val;
+  char *end;
+
+  val = strtoul(str, &end, 0);
+  if ( ! *str || *end ) {
+    fprintf(stderr, "%s: option -%c requires a numeric argument, got %s\n",
+	    program, option, str);
+    exit(1);
+  }
+  return val;
+}
+
+static void usage(void)
+{
+  fprintf(stderr, "Usage: %s [options] path map_type [args...]\n", program);
+}
+
 int main(int argc, char *argv[])
 {
 	char *path, *map, *mapfmt, **mapargv;
 	struct sigaction sa;
-	int mapargc;
-
+	int mapargc, opt;
+	static const struct option long_options[] = {
+	  {"help",    0, 0, 'h'},
+	  {"timeout", 1, 0, 't'},
+	  {"version", 0, 0, 'v'},
+	  {0,0,0,0}
+	};
+	  
 	program = argv[0];
 
-	if ( argc < 3 || geteuid() != 0 ) {
-		fprintf(stderr, "Linux automount version %s\n", version);
-		fprintf(stderr, "Usage: %s path map_type [args...]\n", program);
-		exit(1);
+	ap.exp_timeout = DEFAULT_TIMEOUT;
+
+	opterr = 0;
+	while ( (opt = getopt_long(argc, argv, "+ht:v", long_options,
+				   NULL)) != EOF ) {
+	  switch( opt ) {
+	  case 'h':
+	    usage();
+	    exit(0);
+	  case 't':
+	    ap.exp_timeout = getnumopt(optarg, opt);
+	    break;
+	  case 'v':
+	    printf("Linux automount version %s\n", version);
+	    exit(0);
+	  case '?':
+	  case ':':
+	    printf("%s: Ambiguous or unknown options\n", program);
+	    exit(1);
+	  }
+	}
+	
+	if ( geteuid() != 0 ) {
+	  fprintf(stderr, "%s: This program must be run by root.\n", program);
+	  exit(1);
 	}
 
-	initialize();
+	/* Remove the options */
+	argv += optind;
+	argc -= optind;
+	
+	if ( argc < 2 ) {
+	  usage();
+	  exit(1);
+	}
+	
+	become_daemon();
+	
+	path =    argv[0];
+	map  =    argv[1];
+	mapargv = &argv[2];
+        mapargc = argc-2;         
 
-	path =    argv[1];
-	map  =    argv[2];
-	mapargv = &argv[3];
-        mapargc = argc-3;         
-
-	syslog(LOG_INFO, "starting automounter version %s, path = %s, maptype = %s, mapname = %s", version, path, map, (argc < 4) ? "none" : mapargv[0]);
+	syslog(LOG_INFO, "starting automounter version %s, path = %s, "
+	       "maptype = %s, mapname = %s", version, path, map,
+	       (argc < 3) ? "none" : mapargv[0]);
 
 	if ( (mapfmt = strchr(map,',')) )
 		*(mapfmt++) = '\0';
 
-	if ( !(ap.lookup = open_lookup(map, "", mapfmt, argc-3, mapargv)) )
+	if ( !(ap.lookup = open_lookup(map, "", mapfmt, mapargc, mapargv)) )
 		exit(1);
 
 	if (mount_autofs(path) < 0) {
@@ -529,13 +598,15 @@ int main(int argc, char *argv[])
 	} else {
 		unsigned long timeout;
 
-		ap.exp_timeout = DEFAULT_TIMEOUT;
 		ap.exp_runfreq = ap.exp_timeout / CHECK_RATIO;
 
 		timeout = ap.exp_timeout;
 		ioctl(ap.ioctlfd, AUTOFS_IOC_SETTIMEOUT, &timeout);
 		
-		alarm(ap.exp_timeout+ap.exp_runfreq);
+		/* We often start several automounters at the same
+		   time.  Add some randomness so we don't all expire
+		   at the same time. */
+		alarm(ap.exp_timeout + my_pid % ap.exp_runfreq);
 	}
 
 	while ( !shutdown ) {
