@@ -122,9 +122,9 @@ int mount_mount(char *root, char *name, int name_len, char *what,
   } else {
     /* Not a local host - do a mount */
 
-    fullpath = malloc(strlen(root)+name_len+2);
+    fullpath = alloca(strlen(root)+name_len+2);
     if ( !fullpath ) {
-      syslog(LOG_ERR, MODPREFIX "malloc: %m");
+      syslog(LOG_ERR, MODPREFIX "alloca: %m");
       return 1;
     }
     sprintf(fullpath, "%s/%s", root, name);
@@ -133,7 +133,6 @@ int mount_mount(char *root, char *name, int name_len, char *what,
     syslog(LOG_DEBUG, MODPREFIX "calling mkdir %s", fullpath);
     if ( mkdir(fullpath, 0555) && errno != EEXIST ) {
       syslog(LOG_NOTICE, MODPREFIX "mkdir %s failed: %m", name);
-      free(fullpath);
       return 1;
     }
     
@@ -146,13 +145,15 @@ int mount_mount(char *root, char *name, int name_len, char *what,
       syslog(LOG_DEBUG, MODPREFIX "calling mount -t nfs %s %s", what, fullpath);
       err = spawnl(_PATH_MOUNT, _PATH_MOUNT, "-t", "nfs", what, fullpath, NULL);
     }
-    free(fullpath);
-    if ( err )
-      syslog(LOG_NOTICE, MODPREFIX "nfs entry %s: mount failure",
-	     name);
-    else
-      syslog(LOG_DEBUG, MODPREFIX "mounted %s", name);
-    return ( err != 0 );
+    if ( err ) {
+      rmdir(fullpath);
+      syslog(LOG_NOTICE, MODPREFIX "nfs: mount failure %s on %s",
+	     what, fullpath);
+      return 1;
+    } else {
+      syslog(LOG_DEBUG, MODPREFIX "mounted %s on %s", what, fullpath);
+      return 0;
+    }
   }
 }
 
